/* check.c */

#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <ftw.h>
#include <string.h>

#include <db.h>

#include <sys/types.h>

#include "md5.h"
#include "wa.h"


static DB *dbp;
static int verbose = 0;
static char *database_path = "tmp.db";



static int
check_object(const char *path,
	     const struct stat *sp,
	     int f,
	     struct FTW * fp)
{
    wa_fileinfo_t *fip;
    int err;
    DBT key, data;
    md5_digest_t cur_digest;
    

    if (sp == NULL)
	return 0;
    
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.size = strlen(path)+1;
    key.data = (char *) path;

    err = dbp->get(dbp, NULL, &key, &data, 0);
    if (err)
    {
	printf("%s: New file (not found in database)\n", path);
	return 0;
    }

    fip = data.data;
    
    if (f == FTW_F)
    {
	md5_file(path, &cur_digest);

	if (memcmp(&cur_digest, &fip->digest, sizeof(cur_digest)) != 0)
	    printf("%s: MD5 mismatch\n", path);
    }
    
    return 0;
}


static char *argv0;


static void
usage(FILE *fp)
{
    fprintf(fp, "Usage: %s [-v] DB path [... path]\n", argv0);
}


static int
check_database(void)
{
    int err;
    DBC *cursor;
    DBT key, data;


    
    err = dbp->cursor(dbp, NULL, &cursor, 0);
    if (err)
    {
	fprintf(stderr, "db_cursor: %s\n", strerror(err));
	return -1;
    }

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    

    while ((err = cursor->c_get(cursor, &key, &data, DB_NEXT)) == 0)
    {
	struct stat statbuf;
	md5_digest_t digest;
	wa_fileinfo_t *fip;

	if (verbose)
	    puts(key.data);

	fip = data.data;
	if (fip == NULL)
	{
	    /* Ignore empty records for now */
	    continue;
	}
	
	if (S_ISREG(fip->statbuf.st_mode))
	{
	    if (lstat(key.data, &statbuf) < 0)
	    {
		printf("%s: %s\n", (char *) key.data, strerror(errno));
		continue;
	    }
	    
	    if (!S_ISREG(statbuf.st_mode))
	    {
		printf("%s: Not a file anymore\n", (char *) key.data);
		continue;
	    }
	    
	    md5_file(key.data, &digest);

	    if (memcmp(&digest, &fip->digest, sizeof(digest)) != 0)
		printf("%s: MD5 mismatch\n", (char *) key.data);
	}
    }

    if (err > 0 && err != DB_NOTFOUND)
    {
	fprintf(stderr, "c_get: %s\n", strerror(err));
	err = cursor->c_close(cursor);
    }
    
    err = cursor->c_close(cursor);
    return 0;
}



int
main(int argc,
     char *argv[])
{
    int i, err;
    

    argv0 = argv[0];
    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	  case 'v':
	    verbose = 1;
	    break;

	  default:
	    fprintf(stderr, "%s: %s: invalid command line option\n",
		    argv[0], argv[i]);
	    usage(stderr);
	    exit(1);
	}

    if (i == argc)
    {
	fprintf(stderr, "%s: missing database argument\n", argv[0]);
	exit(1);
    }

    database_path = argv[i++];
    
    err = db_open(database_path, DB_UNKNOWN, DB_RDONLY, 0, NULL, NULL, &dbp);
    if (err)
    {
	fprintf(stderr, "db_open: %s: %s\n", 
		database_path, strerror(err));
	exit(1);
    }

    if (i == argc)
	check_database();
    else
    {
	for (; i < argc; i++)
	    nftw(argv[i], check_object, 200, FTW_PHYS+FTW_MOUNT+FTW_CHDIR);
    }
    
    err = dbp->close(dbp, 0);
    
    exit(0);
}
	
		   
