/* diff.c */

#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <db.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "md5.h"
#include "wa.h"

static DB *dbp1;
static DB *dbp2;

static int verbose = 0;
static int debug = 0;

static char *database1_path = "tmp.db";
static char *database2_path = "tmp2.db";

static char *argv0;


static void
usage(FILE *fp)
{
    fprintf(fp, "Usage: %s [-v] DB1 DB2\n", argv0);
}



int
main(int argc,
     char *argv[])
{
    int i, err, err1, err2, diff;
    DBC *c1, *c2;
    DBT k1, k2;
    DBT d1, d2;
    

    argv0 = argv[0];
    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	  case 'v':
	    verbose = 1;
	    break;

	  case 'd':
	    debug = 1;
	    break;

	  default:
	    fprintf(stderr, "%s: %s: invalid command line option\n",
		    argv[0], argv[i]);
	    usage(stderr);
	    exit(1);
	}


    if (i+2 != argc)
    {
	fprintf(stderr, "%s: missing database arguments\n", argv[0]);
	exit(1);
    }

    database1_path = argv[i++];
    database2_path = argv[i++];
    
    err = db_open(database1_path, DB_UNKNOWN, DB_RDONLY, 0, NULL, NULL, &dbp1);
    if (err)
    {
	fprintf(stderr, "db_open: %s: %s\n", database1_path, strerror(err));
	exit(1);
    }

    /* Open database 2 for RD/WR */
    err = db_open(database2_path, DB_UNKNOWN, DB_RDONLY, 0, NULL, NULL, &dbp2);
    if (err)
    {
	fprintf(stderr, "db_open: %s: %s\n", database2_path, strerror(err));
	exit(1);
    }

	   
    err = dbp1->cursor(dbp1, NULL, &c1, 0);
    if (err)
    {
	fprintf(stderr, "db_cursor: %s\n", strerror(err));
	goto Fail;
    }
    
    err = dbp2->cursor(dbp2, NULL, &c2, 0);
    if (err)
    {
	fprintf(stderr, "db_cursor: %s\n", strerror(err));
	goto Fail;
    }
    
    memset(&k1, 0, sizeof(k1));
    memset(&k2, 0, sizeof(k2));
    memset(&d1, 0, sizeof(d1));
    memset(&d2, 0, sizeof(d2));

    err1 = err2 = 0;
    
    while (1)
    {
        wa_fileinfo_t *f1, *f2;


	if (!err1)
	    err1 = c1->c_get(c1, &k1, &d1, DB_NEXT);
	if (!err2)
	    err2 = c2->c_get(c2, &k2, &d2, DB_NEXT);

      Again:
	if (debug)
	{
	    if (!err1)
		printf("DEBUG: c_get(DB1): %s\n", (char *) k1.data);
	    if (!err2)
		printf("DEBUG: c_get(DB2): %s\n", (char *) k2.data);
	}
	
	if (err1 && err2)
	    break;

	if (err1 && !err2)
	{
	    printf("%s: New\n", (char *) k2.data);
	    continue;
	}

	if (err2 && !err1)
	{
	    printf("%s: Removed\n", (char *) k1.data);
	    continue;
	}
	
	diff = strcmp(k1.data, k2.data);
	if (diff > 0)
	{
	    /* Key is missing in DB1 */
	    printf("%s: New\n", (char *) k2.data);
	    err2 = c2->c_get(c2, &k2, &d2, DB_NEXT);
	    goto Again;
	}
	else if (diff < 0)
	{
	    /* Key is missing in DB2 */
	    printf("%s: Removed\n", (char *) k1.data);
	    err1 = c1->c_get(c1, &k1, &d1, DB_NEXT);
	    goto Again;
	}
	else
	{
	    /* Same key */

	    f1 = d1.data;
	    f2 = d2.data;

	    if (!f1 && !f2)
	    {
		/* XXX: Fixme */
		printf("%s: Huh!?\n", (char *) k1.data);
		continue;
	    }
	    
	    if ((f1 && !f2) || (!f1 && f2) ||
		f1->statbuf.st_mode != f2->statbuf.st_mode ||
		f1->statbuf.st_size != f2->statbuf.st_size ||
		f1->statbuf.st_mtime != f2->statbuf.st_mtime ||
		memcmp(&f1->digest, &f2->digest, sizeof(f1->digest)) != 0)
	    {
		printf("%s: Modified\n", (char *) k1.data);
	    }
	}
    }

    c1->c_close(c1);
    c2->c_close(c2);

    dbp1->close(dbp1, 0);
    dbp2->close(dbp2, 0);
    
    exit(0);
    
  Fail:
    c1->c_close(c1);
    c2->c_close(c2);
    
    dbp1->close(dbp1, 0);
    dbp2->close(dbp2, 0);
    exit(1);
}
	
		   
