/* find.c */

#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <db.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "md5.h"
#include "wa.h"

static int verbose = 0;
static int debug = 0;
static char *database_path = "tmp.db";
static char *index_database_path = NULL;



static int
print_object(const char *path,
	     wa_fileinfo_t *fip)
{
    puts(path);
    return 0;
}


static char *argv0;


static void
usage(FILE *fp)
{
    fprintf(fp, "Usage: %s [-v] DB\n", argv0);
}


int
search_main_database(const char *db_path,
		     const char *str)
{
    int err;
    DB *dbp;
    DBC *c;
    DBT key, data;
    
    
    err = db_open(db_path, DB_UNKNOWN, DB_RDONLY, 0, NULL, NULL, &dbp);
    if (err)
    {
	fprintf(stderr, "db_open: %s: %s\n", database_path, strerror(err));
	return -1;
    }


    err = dbp->cursor(dbp, NULL, &c, 0);
    if (err)
    {
	fprintf(stderr, "db_cursor: %s\n", strerror(err));
	goto Fail;
    }

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    

    while ((err = c->c_get(c, &key, &data, DB_NEXT)) == 0)
    {
	char *cp;
	int len = strlen(str);
	
	cp = strstr(key.data, str);
	if (cp != NULL &&
	    cp[len] == '\0' &&
	    (cp == key.data || cp[-1] == '/'))
	    
	{
	    print_object(key.data, data.data);
	}
    }

    if (err > 0 && err != DB_NOTFOUND)
    {
	fprintf(stderr, "c_get: %s\n", strerror(err));
	goto Fail;
    }

    err = c->c_close(c);
    err = dbp->close(dbp, 0);
    return 0;

  Fail:
    err = c->c_close(c);
    err = dbp->close(dbp, 0);
    return -1;
}    


int
search_index_database(const char *db_path,
		      const char *idb_path,
		      const char *str)
{
    int err;
    DB *dbp, *idbp;
    DBC *c;
    DBT key, data;
    
    
    err = db_open(idb_path, DB_UNKNOWN, DB_RDONLY, 0, NULL, NULL, &idbp);
    if (err)
    {
	fprintf(stderr, "db_open: %s: %s\n", idb_path, strerror(err));
	return -1;
    }

    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    key.data = (void *) str;
    key.size = strlen(str)+1;
    
    err = idbp->cursor(idbp, NULL, &c, 0);
    if (err)
    {
	fprintf(stderr, "cursor: %s\n", strerror(err));
	idbp->close(idbp, 0);
	return -1;
    }
    
    err = c->c_get(c, &key, &data, DB_SET);
    if (err)
    {
	/* Key not found in index database */

	/* TODO: Linear search of the index database? */
	c->c_close(c);
	idbp->close(idbp, 0);

	/* Revert to searching the main database */
	return search_main_database(db_path, str);
    }
    
    if (debug)
	fprintf(stderr, "DEBUG: Found key in index database: %s:%s\n",
		(char *) key.data, (char *) data.data);
    
    err = db_open(db_path, DB_UNKNOWN, DB_RDONLY, 0, NULL, NULL, &dbp);
    if (err)
    {
	fprintf(stderr, "db_open: %s: %s\n", db_path, strerror(err));
	c->c_close(c);
	idbp->close(idbp, 0);
	return -1;
    }

    while (err == 0)
    {
	DBT key2, data2;

	memset(&key2, 0, sizeof(key2));
	memset(&data2, 0, sizeof(data2));
	
	key2 = data;
	
	err = dbp->get(dbp, NULL, &key2, &data2, 0);
	if (err)
	{
	    fprintf(stderr, "%s: Not found in main database\n",
		    (char *) key2.data);
	}
	else
	{
	    print_object(key2.data, data2.data);
	}

	err = c->c_get(c, &key, &data, DB_NEXT_DUP);
    }

    if (err > 0 && err != DB_NOTFOUND)
    {
	fprintf(stderr, "c_get: %s\n", strerror(err));
	goto Fail;
    }

    err = c->c_close(c);
    err = idbp->close(idbp, 0);
    err = dbp->close(dbp, 0);
    return 0;

  Fail:
    err = c->c_close(c);
    err = idbp->close(idbp, 0);
    err = dbp->close(dbp, 0);
    return -1;
}    


int
main(int argc,
     char *argv[])
{
    int i;
    

    argv0 = argv[0];
    
    for (i = 1; i < argc && argv[i][0] == '-'; i++)
	switch (argv[i][1])
	{
	  case 'v':
	    verbose = 1;
	    break;

	  case 'd':
	    debug = 1;
	    break;

	  case 'i':
	    if (argv[i][2])
		index_database_path = argv[i]+2;
	    else
		index_database_path = argv[++i];
	    break;
	    
	  default:
	    fprintf(stderr, "%s: %s: invalid command line option\n",
		    argv[0], argv[i]);
	    usage(stderr);
	    exit(1);
	}

    if (i >= argc)
    {
	fprintf(stderr, "%s: missing database argument\n", argv[0]);
	exit(1);
    }
	
    database_path = argv[i++];

    if (index_database_path)
	search_index_database(database_path,
			      index_database_path,
			      argv[i]);
    else
	search_main_database(database_path,
			     argv[i]);

    return 0;
}
	
		   
